# BEGIN LICENSE BLOCK
# 
# Copyright (c) 1996-2003 Jesse Vincent <jesse@bestpractical.com>
# 
# (Except where explictly superceded by other copyright notices)
# 
# This work is made available to you under the terms of Version 2 of
# the GNU General Public License. A copy of that license should have
# been provided with this software, but in any event can be snarfed
# from www.gnu.org.
# 
# This work is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# Unless otherwise specified, all modifications, corrections or
# extensions to this work which alter its source code become the
# property of Best Practical Solutions, LLC when submitted for
# inclusion in the work.
# 
# 
# END LICENSE BLOCK
use strict;

no warnings qw(redefine);
use vars qw(%_ACL_KEY_CACHE);

use RT::Group;
use RT::User;

# {{{ IsGroup

=head2 IsGroup

Returns true if this principal is a group. 
Returns undef, otherwise

=cut

sub IsGroup {
    my $self = shift;
    if ($self->PrincipalType eq 'Group') {
        return(1);
    }
    else {
        return undef;
    }
}

# }}}

# {{{ IsUser

=head2 IsUser 

Returns true if this principal is a User. 
Returns undef, otherwise

=cut

sub IsUser {
    my $self = shift;
    if ($self->PrincipalType eq 'User') {
        return(1);
    }
    else {
        return undef;
    }
}

# }}}

# {{{ Object

=head2 Object

Returns the user or group associated with this principal

=cut

sub Object {
    my $self = shift;

    unless ($self->{'object'}) {
    if ($self->IsUser) {
       $self->{'object'} = RT::User->new($self->CurrentUser);
    }
    elsif ($self->IsGroup) {
        $self->{'object'}  = RT::Group->new($self->CurrentUser);
    }
    else { 
        $RT::Logger->crit("Found a principal (".$self->Id.") that was neither a user nor a group");
        return(undef);
    }
    $self->{'object'}->Load($self->ObjectId());
    }
    return ($self->{'object'});


}
# }}} 


# {{{ GrantRight 

=head2 GrantRight  { Right => RIGHTNAME, Object => undef }

A helper function which calls RT::ACE->Create

=cut

sub GrantRight {
    my $self = shift;
    my %args = ( Right => undef,
                Object => undef,
                @_);


    #if we haven't specified any sort of right, we're talking about a global right
    if (!defined $args{'Object'} && !defined $args{'ObjectId'} && !defined $args{'ObjectType'}) {
        $args{'Object'} = $RT::System;
    }

    unless ($args{'Right'}) {
        return(0, $self->loc("Invalid Right"));
    }


    #ACL check handled in ACE.pm
    my $ace = RT::ACE->new( $self->CurrentUser );


    my $type = $self->_GetPrincipalTypeForACL();

    # If it's a user, we really want to grant the right to their 
    # user equivalence group
        return ( $ace->Create(RightName => $args{'Right'},
                          Object => $args{'Object'},
                          PrincipalType =>  $type,
                          PrincipalId => $self->Id
                          ) );
}
# }}}

# {{{ RevokeRight

=head2 RevokeRight { Right => "RightName", Object => "object" }

Delete a right that a user has 

=cut

sub RevokeRight {

    my $self = shift;
    my %args = (
        Right      => undef,
        Object => undef,
        @_
    );

    #if we haven't specified any sort of right, we're talking about a global right
    if (!defined $args{'Object'} && !defined $args{'ObjectId'} && !defined $args{'ObjectType'}) {
        $args{'Object'} = $RT::System;
    }
    #ACL check handled in ACE.pm
    my $type = $self->_GetPrincipalTypeForACL();

    my $ace = RT::ACE->new( $self->CurrentUser );
    $ace->LoadByValues(
        RightName     => $args{'Right'},
        Object    => $args{'Object'},
        PrincipalType => $type,
        PrincipalId   => $self->Id
    );

    unless ( $ace->Id ) {
        return ( 0, $self->loc("ACE not found") );
    }
    return ( $ace->Delete );
}

# }}}


# {{{ sub Rights

=head2 sub Rights (Objects => undef)

Check to see what rights this user has for the objects in the arrayref Objects

This still hard codes to check to see if a user has queue-level rights
if we ask about a specific ticket.

Return is unimportant. it just populates the cached.

=cut

sub Rights {

    my $self = shift;
    my %args = (
                 Objects     => undef,
                 @_ );

    if ( $self->Disabled ) {
        $RT::Logger->err( "Disabled User:  " . $self->id . " failed access check for " . $args{'Right'} );
        return (undef);
    }


    # If this object is a ticket, we care about ticket roles and queue roles
    my ( $or_look_at_object_rights, $or_check_roles );

    my @roles;

    foreach my $object (@{$args{'Objects'}}) { 
          push (@roles, $self->_RolesForObject(ref($object), $object->id));
    }


    if (@roles) {
        $or_check_roles = " OR ( (".join (' OR ', @roles)." ) ".  
        " AND Groups.Type = ACL.PrincipalType AND Groups.Id = Principals.id AND Principals.PrincipalType = 'Group') ";

   } 
    # {{{ Construct Right Match

    # If an object is defined, we want to look at rights for that object
   
    my @look_at_objects;



    foreach my $obj (@{$args{'Objects'}}) {
            next unless (UNIVERSAL::can($obj, 'id'));
            my $type = ref($obj);
            my $id = $obj->id;
            push @look_at_objects, "(ACL.ObjectType = '$type' AND ACL.ObjectId = $id)"; 
            }

     
    # }}}


    my $query = "SELECT ACL.RightName, ACL.ObjectType, ACL.ObjectId, CachedGroupMembers.MemberId FROM ACL, Groups, Principals, CachedGroupMembers WHERE  ".
    # Only find superuser or rights with the name $right
   # Never find disabled groups.
   "Principals.Disabled = 0 " .  "AND CachedGroupMembers.Disabled = 0  ".  "AND Principals.id = Groups.id " .  
    # We always grant rights to Groups

    # See if the principal is a member of the group recursively or _is the rightholder_
    # never find recursively disabled group members
    # also, check to see if the right is being granted _directly_ to this principal,
    #  as is the case when we want to look up group rights
    "AND  Principals.id = CachedGroupMembers.GroupId AND CachedGroupMembers.MemberId = " . $self->Id . " ".

    # Make sure the rights apply to the entire system or to the object in question
    "AND ( ".join(' OR ', @look_at_objects).") ".

    # limit the result set to groups of types ACLEquivalence (user)  UserDefined, SystemInternal and Personal
    "AND ( (  ACL.PrincipalId = Principals.id AND ACL.PrincipalType = 'Group' AND ".
        "(Groups.Domain = 'SystemInternal' OR Groups.Domain = 'UserDefined' OR Groups.Domain = 'ACLEquivalence' OR Groups.Domain = 'Personal'))".

        # have a look at role groups, if there are any
         $or_check_roles.
        " ) ";

    #   $RT::Logger->debug("Now Trying $query");	
    my $acl = $self->_Handle->SimpleQuery($query)->fetchall_arrayref({});
 

    my $time = time();
    foreach my $obj (@{$args{'Objects'}}) {
            next unless (UNIVERSAL::can($obj, 'id'));
            my $type = ref($obj);
            my $id = $obj->id;
           $_ACL_KEY_CACHE{$type}->{$id}->{$self->Id}->{'_queried'} = $time; 
    }
    foreach my $row (@$acl) {
           $_ACL_KEY_CACHE{$row->{'ObjectType'}}->{$row->{'ObjectId'}}->{$row->{'MemberId'}}->{$row->{'RightName'}} = $time;
    }
   
 
}

# }}}

# {{{ sub HasRight

=head2 sub HasRight (Right => 'right' Object => undef)


Checks to see whether this principal has the right "Right" for the Object
specified. If the Object parameter is omitted, checks to see whether the 
user has the right globally.

This still hard codes to check to see if a user has queue-level rights
if we ask about a specific ticket.


This takes the params:

    Right => name of a right

    And either:

    Object => an RT style object (->id will get its id)




Returns 1 if a matching ACE was found.

Returns undef if no ACE was found.

=cut

sub HasRight {

    my $self = shift;
    my %args = ( Right      => undef,
                 Object     => undef,
                 EquivObjects    => undef,
                 @_ );


    if ( $self->Disabled ) {
        $RT::Logger->err( "Disabled User:  " . $self->id . " failed access check for " . $args{'Right'} );
        return (undef);
    }

    if ( !defined $args{'Right'} ) {
        require Carp;
        $RT::Logger->debug( Carp::cluck("HasRight called without a right") );
        return (undef);
    }

    if ( defined( $args{'Object'} )) {
        return (undef) unless (UNIVERSAL::can( $args{'Object'}, 'id' ) );
        push(@{$args{'EquivObjects'}}, $args{Object});
    }
    else {
        $RT::Logger->crit("$self HasRight called with no valid object");
        return (undef);
    }

    # If this object is a ticket, we care about ticket roles and queue roles
    if ( (ref($args{'Object'}) eq 'RT::Ticket') && $args{'Object'}->Id) {
        # this is a little bit hacky, but basically, now that we've done the ticket roles magic, we load the queue object
        # and ask all the rest of our questions about the queue.
        push (@{$args{'EquivObjects'}}, $args{'Object'}->QueueObj);

    }


    push(@{$args{'EquivObjects'}}, $RT::System)
        unless ( (ref($args{'Object'}) eq 'RT::System' ) || ( $self->can('_IsOverrideGlobalACL') and $self->_IsOverrideGlobalACL($args{Object})));

    my $cache_timeout = ( time - 60 );
    my $cache_dirty = 0;
    foreach my $obj (@{$args{'EquivObjects'}}) {
        
            next unless (UNIVERSAL::can($obj, 'id'));
            my $type = ref($obj);
            my $id = $obj->id;
                # if this chunk of the ACL cache is too old or non-existent, clean it out.
           if ( !exists $self->_ACLCache->{$type}->{$id}->{$self->Id}->{'_queried'} ||
                  $self->_ACLCache->{$type}->{$id}->{$self->Id}->{'_queried'} < $cache_timeout) {
                delete $self->_ACLCache->{$type}->{$id}->{$self->Id};
                $cache_dirty = 1;

    }
        return (1) if (
           $self->_ACLCache->{$type}->{$id}->{$self->Id}->{$args{'Right'}} ||
           $self->_ACLCache->{'RT::System'}->{'1'}->{$self->Id}->{'SuperUser'});



    }
    # We didn't find anything. and the cache wasn't fresh enough. let's try once more.
    if ($cache_dirty) {
    $self->Rights(Objects => $args{'EquivObjects'});

    foreach my $obj (@{$args{'EquivObjects'}}) {
            next unless (UNIVERSAL::can($obj, 'id'));
            my $type = ref($obj);
            my $id = $obj->id;
        return (1) if (
           $self->_ACLCache->{$type}->{$id}->{$self->Id}->{$args{'Right'}} ||
           $self->_ACLCache->{'RT::System'}->{'1'}->{$self->Id}->{'SuperUser'});

    }

    }
    #    nothing matched.
    return(0);

}
# }}}
# {{{ _RolesForObject



=head2 _RolesForObject( $object_type, $object_id)

Returns an SQL clause finding role groups for Objects

=cut


sub _RolesForObject {
    my $self = shift;
    my $type = shift;
    my $id = shift;
    my $clause = "(Groups.Domain = '".$type."-Role' AND Groups.Instance = '" . $id. "') ";

    return($clause);
}

# }}}

# }}}

# {{{ ACL caching

# {{{ _ACLCache

=head2 _ACLCache

# Function: _ACLCache
# Type    : private instance
# Args    : none
# Lvalue  : hash: ACLCache
# Desc    : Returns a reference to the Key cache hash

=cut

sub _ACLCache {
    return(\%_ACL_KEY_CACHE);
}

# }}}

# {{{ _InvalidateACLCache

=head2 _InvalidateACLCache

Cleans out and reinitializes the user rights key cache

=cut

sub _InvalidateACLCache {
    %_ACL_KEY_CACHE = ();
}

# }}}

# }}}


# {{{ _GetPrincipalTypeForACL

=head2 _GetPrincipalTypeForACL

Gets the principal type. if it's a user, it's a user. if it's a role group and it has a Type, 
return that. if it has no type, return group.

=cut

sub _GetPrincipalTypeForACL {
    my $self = shift;
    my $type;    
    if ($self->PrincipalType eq 'Group' && $self->Object->Domain =~ /Role$/) {
        $type = $self->Object->Type;
    }
    else {
        $type = $self->PrincipalType;
    }

    return($type);
}

# }}}

# {{{ _ReferenceId

=head2 _ReferenceId

Returns a list uniquely representing an object or normal scalar.

For scalars, its string value is returned; for objects that has an
id() method, its class name and Id are returned as a string seperated by a "-".



=cut


# TODO: does _anyone_ use this? autrijus? if not, cut it and I'll yank it on merge

sub _ReferenceId {
    my $scalar = shift;

    # just return the value for non-objects
    return $scalar unless UNIVERSAL::can($scalar, 'id');

    # an object -- return the class and id
    return(ref($scalar)."-". $scalar->id);
}

# }}}

1;
